/*
/*
 *  INJECT utility (c) Copyright, Kevin Thacker 2012
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* This utility can be used to add a AMSDOS header to a file. The file will
 * be treated by the Amstrad operating system as binary. The resulting file
 * can be injected into a disk image.
 *
 * This utility is ideal as part of a cross-development tool.
*/
#include "opth.h"
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
typedef int BOOL;
#ifndef TRUE
#define TRUE (1==1)
#endif

#ifndef FALSE
#define FALSE (1==0)
#endif

/* I am using a enum, so that I can poke data into structures without
worrying how the compiler has aligned it */
enum
{
	CPC_DISC_HEADER_FILENAME_BYTE0 = 0,
	CPC_DISC_HEADER_FILENAME_BYTE1,
	CPC_DISC_HEADER_FILENAME_BYTE2,
	CPC_DISC_HEADER_FILENAME_BYTE3,
	CPC_DISC_HEADER_FILENAME_BYTE4,
	CPC_DISC_HEADER_FILENAME_BYTE5,
	CPC_DISC_HEADER_FILENAME_BYTE6,
	CPC_DISC_HEADER_FILENAME_BYTE7,
	CPC_DISC_HEADER_FILENAME_BYTE8,
	CPC_DISC_HEADER_FILENAME_EXTENSION0,
	CPC_DISC_HEADER_FILENAME_EXTENSION1,
	CPC_DISC_HEADER_FILENAME_EXTENSION2,
	CPC_DISC_HEADER_FILE_TYPE = 18,
	CPC_DISC_HEADER_LENGTH_LOW,
	CPC_DISC_HEADER_LENGTH_HIGH,
	CPC_DISC_HEADER_LOCATION_LOW,
	CPC_DISC_HEADER_LOCATION_HIGH,
	CPC_DISC_HEADER_FIRST_BLOCK_FLAG,
	CPC_DISC_HEADER_LOGICAL_LENGTH_LOW,
	CPC_DISC_HEADER_LOGICAL_LENGTH_HIGH,
	CPC_DISC_HEADER_EXECUTION_ADDRESS_LOW,
	CPC_DISC_HEADER_EXECUTION_ADDRESS_HIGH,
	CPC_DISC_HEADER_DATA_LENGTH_LOW = 64,
	CPC_DISC_HEADER_DATA_LENGTH_MID,
	CPC_DISC_HEADER_DATA_LENGTH_HIGH,
	CPC_DISC_HEADER_CHECKSUM_LOW,
	CPC_DISC_HEADER_CHECKSUM_HIGH
} CPC_DISC_HEADER_ENUM;

static int AMSDOSHeader_AllZeros(const char *pData)
{
  unsigned char ordata = 0;
  int i;
    	for (i=0; i<69; i++)
	{
    unsigned char data = pData[i];
    ordata|=data;
	}

    if (ordata)
        return 0;
    return 1;
}

/*--------------------*/
/* calculate checksum */
static unsigned short AMSDOSHeader_CalculateChecksum(const char *pData)
{
	unsigned short CalculatedChecksum = 0;
	unsigned long i;

	/* generate checksum */
	for (i=0; i<66; i++)
	{
    unsigned char data = pData[i];
		CalculatedChecksum+=(unsigned short)(data&0x0ff);
	}
	return CalculatedChecksum;
}

/*------------------------------------------------*/
/* Detect if there is a AMSDOS header on the file */
static int AMSDOSHeader_Checksum(const char *pData)
{
  if (AMSDOSHeader_AllZeros(pData))
  {
    return 0;
  }
else
{  
  
	unsigned short StoredChecksum;
	unsigned short CalculatedChecksum;

	CalculatedChecksum = AMSDOSHeader_CalculateChecksum(pData);

	/* get the stored checksum */
	StoredChecksum = (unsigned short)((pData[67]&0x0ff)|((pData[68]&0x0ff)<<8));

	return (CalculatedChecksum==StoredChecksum);
}
}

/* buffer to hold generated header data */
static const char *InjectFilename;
static BOOL bInjectFileSet=FALSE;
static int nOffset = 0;
static int nStart = 0;
static BOOL bStartSet = FALSE;
static BOOL bOffsetSet=FALSE;

int ReadNumberParameter(const char *param)
{
    int Length = strlen(param);
    BOOL bIsHex = FALSE;
    int Offset = 0;
    unsigned long Value = 0;
    char ch;

    if (Length==0)
        return 0;

    /* check for common prefixs for hex numbers */
    if ((Length>1) && ((param[0]=='&') || (param[0]=='$')))
    {
        Offset = 1;
        bIsHex = TRUE;
    }
    else if ((Length>2) && (param[0]=='0') && ((param[1]=='x') || (param[1]=='X')))
    {
        Offset = 2;
        bIsHex = TRUE;
    }

    if (!bIsHex)
    {
        return atoi(param);
    }

    ch = param[Offset];
    while (ch!='\0')
    {
        Value = Value<<4;
        if ((ch>='0') && (ch<='9'))
        {
            Value = Value | (ch-'0');
        }
        else if ((ch>='a') && (ch<='f'))
        {
            Value = Value | ((ch-'a')+10);
        }
        else if ((ch>='A') && (ch<='F'))
        {
            Value = Value | ((ch-'A')+10);
        }
        Offset++;
        ch = param[Offset];
    }

    return Value;
}

static int NumFiles = 0;
static const char *Filenames[2];

int	NonOptionHandler(const char *pOption)
{
	if (NumFiles<2)
	{
		Filenames[NumFiles] = pOption;
		NumFiles++;
	}

	return OPTION_OK;
}

int SetInjectFilename(ARGUMENT_DATA *pData)
{
	InjectFilename = ArgumentList_GetNext(pData);

	if (InjectFilename==NULL)
		return OPTION_MISSING_PARAMETER;

  bInjectFileSet = TRUE;
  
	return OPTION_OK;
}


int SetOffsetOption(ARGUMENT_DATA *pData)
{
	const char *pAddress = ArgumentList_GetNext(pData);

	if (pAddress==NULL)
		return OPTION_MISSING_PARAMETER;

	nOffset = ReadNumberParameter(pAddress);
	printf("Offset is %d (&%04x)\n", nOffset, nOffset&0x0ffff);
	bOffsetSet = TRUE;

	return OPTION_OK;
}


int SetStartOption(ARGUMENT_DATA *pData)
{
	const char *pAddress = ArgumentList_GetNext(pData);

	if (pAddress==NULL)
		return OPTION_MISSING_PARAMETER;

	nStart = ReadNumberParameter(pAddress);
	printf("Start is %d (&%04x)\n", nStart, nStart&0x0ffff);

	bStartSet = TRUE;

	return OPTION_OK;
}

OPTION OptionTable[]=
{
	{"o",SetOffsetOption},
	{"s",SetStartOption},
	{"i",SetInjectFilename},
	{NULL, NULL},
};


void	OutputDetails(void)
{
	fprintf(stdout,"INJECT\n");
	fprintf(stdout,"\n");
	fprintf(stdout,"(c) Kevin Thacker 2010\n");
	fprintf(stdout,"A utility to inject a file into another file.\n");
	fprintf(stdout,"\n");
	fprintf(stdout,"Usage: INJECT -o <offset> -i <inject filename> <input filename> <output filename>\n");
	fprintf(stdout,"\n");
	fprintf(stdout,"\n");
}

int	main(int argc, char *argv[])
{
	int nReturnCode= 0;

	bOffsetSet = FALSE;
  bStartSet = FALSE;
	bInjectFileSet = FALSE;
  
	NumFiles = 0;

	if (ArgumentList_Execute(argc, argv, OptionTable, printf, NonOptionHandler)==OPTION_OK)
	{
			FILE *fh_in, *fh_out;
			unsigned long filesize;
      printf("INJECT utility\n\n");
    
      if (!bOffsetSet)
      {
          printf("Offset must be set!\n");
        return 1;
      } /* bOffsetSet */

      if (bStartSet)
      {
          printf("Adjusting offset based on start value.\n");
        nOffset-=nStart;
      }
      if (!bInjectFileSet)
      {
          printf("Inject file must be set!\n");
          return 1;
      } /* bInjectFileSet */
      
      printf("Opening input file: %s\n", Filenames[0]);

			/* open input binary file */
			fh_in = fopen(Filenames[0],"rb");

			if (fh_in==NULL)
			{
				printf("Failed to open input file\n");
				nReturnCode = 1;
			}
			else
			{
				unsigned char *file_data;

				/* seek to end of file */
				fseek(fh_in, 0, SEEK_END);

				/* report position in file */
				filesize = ftell(fh_in);

				/* seek to beginning of file */
				fseek(fh_in, 0,SEEK_SET);
  
				file_data = (unsigned char *)malloc(filesize);

				if (file_data!=NULL)
				{
					fread(file_data, 1, filesize, fh_in);
				}
				else
				{
					printf("Failed to allocate memory for input file.\n");
					nReturnCode = 1;
				}
        printf("Input file size: %d (0x%04x)\n", filesize, filesize);
				fclose(fh_in);

        if (nOffset>filesize)
        {
            printf("Bad offset, beyond end of file!n");
          nReturnCode = 4;
        }
        else
        {
          
          if (file_data!=NULL)
          {
            BOOL bSourceHasHeader = FALSE;
            if (AMSDOSHeader_Checksum((const char *)file_data))
            {
              printf("AMSDOS header detected for input file.\n");
              bSourceHasHeader = TRUE;
            } /* AMSDOSHeader_Checksum */

            if (bSourceHasHeader)
            {
                nOffset+=128;
            }
            printf("Opening inject file: %s\n", InjectFilename);
            
            fh_in = fopen(InjectFilename,"rb");
                       
          if (fh_in==NULL)
          {
            printf("Failed to open file to inject\n");
            nReturnCode = 1;
          }
          else
          {
            unsigned char *inject_file_data;
          unsigned long inject_file_size;
              unsigned long write_length;
                BOOL bInjectFileHasHeader= FALSE;
            unsigned long nSourceOffset = 0;
            
            /* seek to end of file */
            fseek(fh_in, 0, SEEK_END);

            /* report position in file */
            inject_file_size = ftell(fh_in);

            /* seek to beginning of file */
            fseek(fh_in, 0,SEEK_SET);

            inject_file_data = (unsigned char *)malloc(inject_file_size);

            if (inject_file_data!=NULL)
            {
              if (fread(inject_file_data, 1, inject_file_size, fh_in)!=inject_file_size)
	      {
      		    printf("Failed to read inject file\n");

		      nReturnCode = 2;
		      free(inject_file_data);
		      inject_file_data = NULL;
	      }
            }
            else
            {
              printf("Failed to allocate memory for inject file.\n");
              nReturnCode = 1;
            }
          printf("Inject file size: %d (0x%04x)\n", inject_file_size, inject_file_size);

            fclose(fh_in);
            
	    
	    
            if (inject_file_data!=NULL)
            {
              if (AMSDOSHeader_Checksum((const char *)inject_file_data))
              {
                printf("AMSDOS header detected for inject file.\n");
                bInjectFileHasHeader = TRUE;
              }

              if (bInjectFileHasHeader)
              {
                  nSourceOffset+=128;
                  inject_file_size-=128;
              }
                    
            write_length = inject_file_size;
            
            if ((inject_file_size+nOffset)>filesize)
            {
              printf("Warning: Injected file will write past end of input file. Truncating");
              write_length = filesize-nOffset;
            }
          printf("Writing injected file to offset: &%04x\n", nOffset);
            
            /* copy it */
            memcpy(file_data+nOffset, inject_file_data+nSourceOffset, write_length);

              /* free inject file */
            free(inject_file_data);
          }   /* inject_file_data!=NULL */
          
        printf("Opening output file: %s\n", Filenames[1]);

          /* now open output file */
          fh_out = fopen(Filenames[1],"wb");

          if (fh_out==NULL)
          {
            printf("Failed to open output file.\n");
            nReturnCode = 1;
          }
          else
          {
            /* copy remaining data to output */
            if (fwrite(file_data, 1, filesize, fh_out)!=filesize)
	    {	
		    printf("Failed to write output file\n");
		    nReturnCode = 2;
	    }

            /* close output file */
            fclose(fh_out);

            printf("Output file written.\n");
          } /* fh_out */
        } /* fh_in inject */
      }
      free(file_data);
    } /* file_data */ 
  }
}/* ArgumentList_Execute */
return nReturnCode;
}

